@@ -63,7 +63,7 @@ T_FOLDER = os.path.join(APP_FOLDER, "translations")
 
 # Celery settings
 USE_CELERY = False
-CELERY_BROKER = 'redis://localhost:6379/0'
+CELERY_BROKER = "redis://localhost:6379/0"
 
 # try import private settings
 try:
